"""

"""

import configparser
import odo
import os
import pandas as pd
import cytominer_database.ingest
import pkg_resources
import subprocess
import tempfile


def test_seed(dataset):
    munge_file = pkg_resources.resource_filename(__name__, "../cytominer_database/scripts/munge.sh")

    if dataset["munge"]:
        subprocess.call([munge_file, dataset["data_dir"]])

    config_file = os.path.join(dataset["data_dir"], "config.ini")

    config = configparser.ConfigParser()

    config.read(config_file)

    with tempfile.TemporaryDirectory() as temp_dir:
        sqlite_file = os.path.join(temp_dir, "test.db")

        cytominer_database.ingest.seed(config=config, source=dataset["data_dir"], target="sqlite:///{}".format(str(sqlite_file)))

        for (k, v) in dict({"cells": "Cells.csv", "cytoplasm": "Cytoplasm.csv", "nuclei": "Nuclei.csv"}).items():
            config["filenames"][k] = v

        for table_key in ["image", "cells", "cytoplasm", "nuclei"]:
            csv_filename = os.path.join(temp_dir, config["filenames"][table_key])

            table_name = config["filenames"][table_key].split(".")[0]

            odo.odo("sqlite:///{}::{}".format(str(sqlite_file), table_name), csv_filename)

            df = pd.read_csv(csv_filename)

            assert df.shape[0] == dataset["ingest"]["{}_nrows".format(table_name)]

            assert df.shape[1] == dataset["ingest"]["{}_ncols".format(table_name)] + 1

            if table_key != "image":
                assert df.groupby(["TableNumber", "ImageNumber"]).size().sum() == dataset["ingest"]["{}_nrows".format(table_name)]



"""``ingest`` - mechanism to ingest CSV files into a database.

    In morphological profiling experiments, a CellProfiler pipeline is often run in parallel across multiple images and
    produces a set of CSV files. For example, imaging a 384-well plate, with 9 sites per well, produces 384 * 9 images;
    a CellProfiler process may be run on each image, resulting in a 384*9 output directories (each directory typically
    contains one CSV file per compartment (e.g. Cells.csv, Cytoplasm.csv, Nuclei.csv) and one CSV file for per-image
    measurements (e.g. Image.csv).

    ``ingest`` can be used to read all these CSV files into a database backend. SQLite is the recommended engine, but ingest
    will likely also work with PostgreSQL and MySQL.

    ``ingest`` assumes a directory structure like shown below:

    | plate_a/
    |   set_1/
    |       file_1.csv
    |       file_2.csv
    |       ...
    |       file_n.csv
    |   set_2/
    |       file_1.csv
    |       file_2.csv
    |       ...
    |       file_n.csv
    |   ...
    |   set_m/
    |       file_1.csv
    |       file_2.csv
    |       ...
    |       file_n.csv

    :Example:

        ``$ ingest plate_a -o sqlite:///plate_a.sqlite -c ingest_config.ini``


"""

import click
import configparser
import csv
import hashlib
import odo
import os
import cytominer_database.utils
import pkg_resources
import subprocess
import tempfile
import sqlalchemy

def __format__(name, header):
    if header in ["ImageNumber", "ObjectNumber"]:
        return header

    return "{}_{}".format(name, header)


def into(input, output, name, identifier):
    """Ingest a CSV file into a table in a database.

    :param input: Input CSV file.
    :param output: Connection string for the database.
    :param name: Table in database into which the CSV file will be ingested
    :param identifier: Unique identifier for ``input``.

    :return:

    """

    with tempfile.TemporaryDirectory() as directory:
        source = os.path.join(directory, os.path.basename(input))

        with open(input, "r") as fin, open(source, "w") as fout:
            reader = csv.reader(fin)
            writer = csv.writer(fout)

            headers = next(reader)
            headers = [__format__(name, header) for header in headers]
            headers = ["TableNumber"] + headers

            writer.writerow(headers)

            [writer.writerow([identifier] + row) for row in reader]

        odo.odo(source, "{}::{}".format(output, name), has_header=True, delimiter=",")


def seed(source, target, config):
    """
    Call functions to create backend

    :param config: Configuration file.
    :param source: Directory containing subdirectories which contain CSV files.
    :param target: Connection string for the database.

    :return:

    """

    directories = sorted(list(cytominer_database.utils.find_directories(source)))

    for directory in directories:
        try:
            patterns, image = cytominer_database.utils.validate_csv_set(config, directory)
        except IOError as e:
            click.echo(e)

            continue

        with open(image, "rb") as document:
            identifier = hashlib.md5(document.read()).hexdigest()

        name, _ = os.path.splitext(config["filenames"]["image"])

        try:
            into(input=image, output=target, name=name, identifier=identifier)
        except sqlalchemy.exc.DatabaseError as e:
            click.echo(e)

            continue

        for pattern in patterns:
            name, _ = os.path.splitext(os.path.basename(pattern))

            into(input=pattern, output=target, name=name, identifier=identifier)


@click.command()
@click.argument(
    "source",
    type=click.Path(exists=True)
)
@click.option(
    "-c",
    "--config_file",
    default=pkg_resources.resource_filename(__name__, os.path.join("config", "config_htqc.ini")),
    type=click.Path(exists=True)
)
@click.option(
    "--munge/--no-munge",
    default=True
)
@click.option(
    "-o",
    "--target",
    type=click.Path(writable=True)
)
@click.version_option(
    version=pkg_resources.require("cytominer_database")[0].version
)
def __main__(config_file, source, target, munge):
    """

    :param config_file: Configuration file.
    :param source: Directory containing subdirectories which contain CSV files.
    :param target: Connection string for the database.
    :param munge: True if the CSV files for individual compartments have been merged into a single CSV file. In this
    case, the CSV will be split into one CSV per compartment.


    :return:

    """

    if munge:
        subprocess.call([pkg_resources.resource_filename(__name__, os.path.join("scripts", "munge.sh")), source])

    config = configparser.ConfigParser()

    config.read(config_file)

    seed(source, target, config)

if __name__ == "__main__":
    __main__()



import setuptools
import setuptools.command.test
import sys


class Test(setuptools.command.test.test):
    user_options = [
        ("pytest-args=", "a", "Arguments to pass to py.test")
    ]

    def initialize_options(self):
        setuptools.command.test.test.initialize_options(self)

        self.pytest_args = []

    def finalize_options(self):
        setuptools.command.test.test.finalize_options(self)

        self.test_args = []

        self.test_suite = True

    def run_tests(self):
        import pytest

        errno = pytest.main(self.pytest_args)

        sys.exit(errno)


setuptools.setup(
    name='cytominer_database',
    version='0.0.0',
    author=[
        'Allen Goodman',
        "Claire McQuin",
        "Shantanu Singh"
    ],
    author_email=[
        "allen.goodman@icloud.com",
        "mcquincl@gmail.com",
        "shsingh@broadinstitute.org"
    ],
    cmdclass={
        'test': Test
    },
    package_data={
        'cytominer_database': [
            'config/*.ini',
            'config/*.json',
            'config/*.sql',
            'scripts/*.sh'
        ],
    },
    packages=setuptools.find_packages(
        exclude=[
            'test',
            'doc'
        ]
    ),
    include_package_data=True,
    install_requires=[
        'click',
        'csvkit',
        'odo',
        'pandas',
        'pytest'
    ],
    entry_points={
        'console_scripts': [
            'ingest=cytominer_database.ingest:__main__',
        ]
    },
    license='BSD',
    url='https://github.com/cytomining/cytominer-database'
)



"""
"""

import glob
import logging
import os
import subprocess

logger = logging.getLogger(__name__)

def find_directories(directory):
    """List subdirectories.

    :param directory: directory

    :return: list of subdirectories of ``directory``

    """
    directories = set()

    filenames = glob.glob(os.path.join(directory, '*/'))

    for filename in filenames:
        pathname = os.path.relpath(filename)

        directories.add(pathname)

    return directories


def validate_csv(csvfile):
    """Validate CSV file.

    The CSV file typically corresponds to either a measurement made on a compartment, e.g. Cells.csv, or on an image,
    e.g. Image.csv. The validation performed is generic - it simply checks for malformed CSV files.

    This uses csvclean to check for validity of a CSV file.

    :param csvfile: CSV file to validate

    :return: True if valid, False otherwise.

    """
    cmd = "csvclean -n {}".format(csvfile)

    try:
        csvcheck = subprocess.check_output(cmd, shell=True)

    except subprocess.CalledProcessError as e:
        if e.returncode == 1:
            return False

        elif e.returncode == 127:
            logger.warning("csvclean not found. Validation will not be rigorous.")
            
        else:
            raise subprocess.CalledProcessError(e.returncode, e.cmd)

    nrows = sum(1 for line in open(csvfile)) - 1

    file_size = os.stat(csvfile).st_size

    return (file_size > 0) & (nrows >= 1) & (csvcheck == b'No errors.\n')


def validate_csv_set(config, directory):
    """Validate a set of CSV files.

    This function validates a set of CSV files in a directory. These CSV files  correspond to measurements
    made on different compartment, e.g. Cells.csv, Cytoplasm.csv, Nuclei.csv. An Image.csv file, corresponding to
    measurements made on the whole image, along with metadata, is also typically present.

    :param config: configuration file - this contains the set of CSV files to validate.
    :param directory: directory containing the CSV files.

    :return: a tuple where the first element is the list of pattern CSV files, the second is the image CSV file.

    """
    image_csv = os.path.join(directory, config["filenames"]["image"])

    if not os.path.isfile(image_csv):
        raise OSError("{} not found in {}. Skipping.".format(config["filenames"]["image"], directory))

    pattern_csvs = [filename for filename in glob.glob(os.path.join(directory, '*.csv')) if filename not in [
        os.path.join(directory, config['filenames']['image']),
        os.path.join(directory, config['filenames']['object']),
        os.path.join(directory, config['filenames']['experiment'])
    ]]

    filenames = pattern_csvs + [image_csv]

    file_checks = dict({(filename, validate_csv(filename)) for filename in filenames})

    if not all(file_checks.values()):
        invalid_files = ",".join([os.path.basename(filename) for (filename, valid) in file_checks.items() if not valid])
        raise OSError("Some files were invalid: {}. Skipping {}.".format(invalid_files, directory))

    return pattern_csvs, image_csv
